import os
import shutil
from tqdm.auto import tqdm
from typing import List

from unibox.utils.uni_traverser import UniTraverser


class DirMoverCopier:
    """
    Directory mover/copier class to perform file moving or copying based on given criteria.

    Attributes:
        src_dir: Source directory.
        dst_dir: Destination directory.
        keep_structure: Whether to keep the directory structure.
        include_extensions: Optional list of file extensions to include.
        exclude_extensions: Optional list of file extensions to exclude.
        action: Action to perform ('move' or 'copy').
    """

    def __init__(self, src_dir: str, dst_dir: str, keep_structure: bool, include_extensions: List[str] = None,
                 exclude_extensions: List[str] = None, action: str = 'move'):
        self.src_dir = src_dir
        self.dst_dir = dst_dir
        self.keep_structure = keep_structure
        self.include_extensions = include_extensions
        self.exclude_extensions = exclude_extensions
        self.action = action

        self._print_debug_string()

    def _print_debug_string(self):
        debug_string = f"Processing with action={self.action}: " \
                       f"src_dir={self.src_dir}, dst_dir={self.dst_dir}, " \
                       f"keep_structure={self.keep_structure}, include={self.include_extensions}, " \
                       f"exclude={self.exclude_extensions}"
        print(debug_string)

    def should_include_file(self, file_name):
        extension = os.path.splitext(file_name)[1]
        if self.include_extensions and extension not in self.include_extensions:
            return False
        if self.exclude_extensions and extension in self.exclude_extensions:
            return False
        return True

    def process_files(self):
        """
        Processes the files in the source directory based on the given criteria and action.
        """
        processed_files = 0

        def file_handler(src_file_path: str):
            nonlocal processed_files
            filename = os.path.basename(src_file_path)
            if self.should_include_file(filename):
                dir_path = os.path.dirname(src_file_path)
                dst_file_dir = self.dst_dir if not self.keep_structure else \
                    os.path.join(self.dst_dir, os.path.relpath(dir_path, self.src_dir))
                os.makedirs(dst_file_dir, exist_ok=True)
                dst_file_path = os.path.join(dst_file_dir, filename)
                if self.action == 'move':
                    shutil.move(src_file_path, dst_file_path)
                    processed_files += 1
                elif self.action == 'copy':
                    shutil.copy2(src_file_path, dst_file_path)
                    processed_files += 1

        traverser = UniTraverser(self.src_dir, include_extensions=self.include_extensions,
                                 exclude_extensions=self.exclude_extensions)
        traverser.traverse(file_handler)

        print(f"Processed {processed_files} files")



import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from urllib.parse import urlparse

import boto3
from omegaconf import OmegaConf
from tqdm.auto import tqdm
import pandas as pd

def parse_s3_url(url: str):
    parsed_url = urlparse(url)
    if parsed_url.scheme != "s3":
        raise ValueError(
            "Expecting 's3' scheme, got: {} in {}.".format(parsed_url.scheme, url)
        )
    return parsed_url.netloc, parsed_url.path.lstrip("/")


def parse_config():
    # change config.yaml with "scripts/config.yaml"
    # if you want to run this script locally
    config = OmegaConf.load("config.yaml")
    return config


class S3Client:
    def __init__(self) -> None:
        # config = parse_config()
        self.s3 = boto3.client(
            "s3",
            # aws_access_key_id=config.aws_access_key_id,
            # aws_secret_access_key=config.aws_secret_access_key,
            # region_name=config.aws_region_name,
        )

    def download(self, s3_uri, target_dir):
        bucket, key = parse_s3_url(s3_uri)
        filename = os.path.basename(s3_uri)
        path = os.path.join(target_dir, filename)
        self.s3.download_file(bucket, key, path)
        return path

    def upload(self, file_path, s3_dir):
        filename = os.path.basename(file_path)
        s3_dir = os.path.join(s3_dir, filename)
        bucket, key = parse_s3_url(s3_dir)
        self.s3.upload_file(file_path, bucket, key)

    def exists(self, s3_uri):
        """Check if a file exists in S3 at the given URI."""
        bucket, key = parse_s3_url(s3_uri)
        try:
            self.s3.head_object(Bucket=bucket, Key=key)
            return True
        except self.s3.exceptions.ClientError as e:
            # The exception is thrown when the object does not exist
            return False

    def download_parallel(self, s3_uri_list, target_dir):
        failed_downloads = []

        def download_one(target_dir, s3_uri):
            self.download(s3_uri, target_dir)

        func = partial(download_one, target_dir)
        with tqdm(desc="Downloading images from S3", total=len(s3_uri_list)) as bar:
            with ThreadPoolExecutor(max_workers=32) as executor:
                # Using a dict for preserving the downloaded file for each future,
                # to store it as a failure if we need that
                futures = {
                    executor.submit(func, file_to_download): file_to_download
                    for file_to_download in s3_uri_list
                }
                for future in as_completed(futures):
                    if future.exception():
                        failed_downloads.append(futures[future])
                    bar.update(1)

        return failed_downloads

    def upload_parallel(self, file_list, target_s3_uri):
        failed_uploads = []

        def upload_one(s3_uri, file_path):
            self.upload(file_path, s3_uri)

        func = partial(upload_one, target_s3_uri)
        with tqdm(desc="Uploading images to S3", total=len(file_list)) as bar:
            with ThreadPoolExecutor(max_workers=32) as executor:
                futures = {
                    executor.submit(func, file_to_upload): file_to_upload
                    for file_to_upload in file_list
                }
                for future in as_completed(futures):
                    if future.exception():
                        failed_uploads.append(futures[future])
                    bar.update(1)

        return failed_uploads

    def walk(self, s3_uri:str):
        bucket, key = parse_s3_url(s3_uri)
        paginator = self.s3.get_paginator("list_objects_v2")
        pages = paginator.paginate(Bucket=bucket, Prefix=key)
        for page in pages:
            for obj in page["Contents"]:
                yield {
                    "key": obj["Key"],
                    "size": obj["Size"],
                    "last_modified": obj["LastModified"],
                    "etag": obj["ETag"],
                    "storage_class": obj["StorageClass"],
                }

    def traverse(self, s3_uri, include_extensions=None, exclude_extensions=None, 
                 relative_unix=False, debug_print=True):
        """
        Traverse through a s3 directory and return direct entries (folder or file) under the directory.

        :param include_extensions: list of extensions to include (e.g. ['.jpg', '.png']), defaults to everything
        :param exclude_extensions: list of extensions to exclude (e.g. ['.txt', '.json']), defaults to nothing
        :param relative_unix: whether to give a relative path (relative to bucket) or not 
        :param debug_print: whether to print debug statuses (e.g., tqdm bar) or not
        """
        bucket, prefix = parse_s3_url(s3_uri)

        if not prefix.endswith('/'):
            prefix += '/'

        paginator = self.s3.get_paginator("list_objects_v2")
        response_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix, Delimiter='/')

        all_entries = []

        if debug_print:
            response_iterator = tqdm(response_iterator, desc="Traversing S3", unit="page")

        for page in response_iterator:
            # Add subdirectories to the list
            for d in page.get('CommonPrefixes', []):
                dir_key = d['Prefix']
                dir_entry = dir_key if relative_unix else f"s3://{bucket}/{dir_key}"
                all_entries.append(dir_entry)

            # Add files to the list, applying filters if specified
            for obj in page.get('Contents', []):
                file_key = obj["Key"]
                if file_key == prefix:  # Skip the input directory itself
                    continue
                
                if (include_extensions is None or any(file_key.endswith(ext) for ext in include_extensions)) and \
                   (exclude_extensions is None or not any(file_key.endswith(ext) for ext in exclude_extensions)):
                    file_entry = file_key[len(prefix):] if relative_unix else f"s3://{bucket}/{file_key}"
                    all_entries.append(file_entry)

        return all_entries


if __name__ == '__main__':
    client = S3Client()
    # s3_uri = "s3://dataset-artstation-uw2/artists/_angelaramos_/"
    s3_uri = "s3://dataset-ingested/gallery-dl/_todo_lists/"
    res = client.traverse(s3_uri)

    print(res)
    print("D")




import pandas as pd
from collections import Counter
from typing import Any, Union
import json  # Import the json module


class CompactJSONEncoder(json.JSONEncoder):
    """Custom JSON Encoder for specific formatting of dictionaries and lists."""
    def iterencode(self, o, _one_shot=False):
        """Overridden method for custom JSON formatting."""
        if isinstance(o, (list, dict)):
            items = []
            if isinstance(o, dict):
                for key, value in o.items():
                    formatted_value = json.dumps(value, separators=(',', ': ')).replace('"', "'")
                    items.append(f"'{key}': {formatted_value}")
            else:  # For lists
                items = [json.dumps(item, separators=(',', ': ')).replace('"', "'") for item in o]
            yield from self._encode_line(items, isinstance(o, dict))
        else:
            yield from super().iterencode(o, _one_shot)

    def _encode_line(self, items, is_dict):
        """Helper method to yield formatted lines."""
        yield '{\n' if is_dict else '[\n'
        for i, item in enumerate(items):
            separator = ',' if i < len(items) - 1 else ''
            yield f'    {item}{separator}\n'
        yield '}\n' if is_dict else ']\n'

class UniPeeker:
    """Utility class for peeking into data with efficient methods."""

    def __init__(self, n: int = 3, console_print: bool = False):
        self.n = n
        self.console_print = console_print

    def peeks(self, data: Any, n: int = None, console_print: bool = None) -> dict:
        """Peek into the data and return metadata and a preview of the data, with efficient handling for large data."""
        peek_n = n if n else self.n
        _print = console_print if console_print is not None else self.console_print

        data_type = type(data).__name__
        meta_dict = {}
        preview = None

        if data_type == 'dict':
            meta_dict, preview = self._peek_dict(data, peek_n)
        elif data_type == 'DataFrame':
            meta_dict, preview = self._peek_dataframe(data, peek_n)
        elif data_type == 'list':
            meta_dict, preview = self._peek_list(data, peek_n)
        # Handling for other data types...

        if _print:
            self._print_info(data_type, meta_dict, preview)

        return {'metadata': meta_dict, 'preview': preview}

    def _peek_dict(self, data: dict, n: int) -> tuple:
        """Peek into a dictionary efficiently."""
        first_n = [(k, data[k]) for k in list(data)[:n]]
        value_types = Counter([type(v).__name__ for v in data.values()])
        is_nested = any(isinstance(v, dict) for v in data.values())

        meta_dict = {
            'len': len(data),
            'value_types': dict(value_types),
            'is_nested': is_nested
        }
        return meta_dict, first_n

    def _peek_dataframe(self, data: pd.DataFrame, n: int) -> tuple:
        """Peek into a DataFrame."""
        meta_dict = {
            'len': len(data),
            'columns': list(data.columns),
            'dtypes': data.dtypes.to_dict(),
            'shape': data.shape
        }
        preview = data.head(n)
        return meta_dict, preview

    def _peek_list(self, data: list, n: int) -> tuple:
        """Peek into a list."""
        first_n = data[:n]
        meta_dict = {
            'len': len(data),
            'item_type': type(data[0]).__name__ if data else 'None'
        }
        return meta_dict, first_n

    def _print_info(self, data_type: str, meta_dict: dict, preview: Any) -> None:
        """Print information about the data using custom pretty print for the metadata."""
        data_len = meta_dict['len']

        # Convert dtypes to string if the data is a DataFrame
        if data_type == 'DataFrame':
            meta_dict['dtypes'] = {k: str(v) for k, v in meta_dict['dtypes'].items()}

        pretty_meta_dict = json.dumps(meta_dict, cls=CompactJSONEncoder, indent=4, sort_keys=True)
        print(f"[{data_type}] of size [{data_len}]:\n{pretty_meta_dict}\n")
        print(f"[Preview]\n{preview}")


if __name__ == '__main__':
    # Example usage:
    # some_dict_or_dataframe = {"name": "John", "age": 30}

    # df
    some_dict_or_dataframe = pd.DataFrame({'A': [1, 2], 'B': [3, 4]})

    unipeeker = UniPeeker()
    result = unipeeker.peeks(some_dict_or_dataframe)

    print("D")



import os
import click
from typing import List

from unibox import UniTraverser
# Assuming UniTraverser is already defined above

class FileRenamer:
    """
    File renamer class to perform file renaming based on given criteria.

    Attributes:
        root_dir: Root directory.
        suffix: The suffix to add to the file names.
        include_extensions: Optional list of file extensions to include.
        exclude_extensions: Optional list of file extensions to exclude.
    """

    def __init__(self, root_dir: str, suffix: str, include_extensions: List[str] = None,
                 exclude_extensions: List[str] = None):
        self.root_dir = root_dir
        self.suffix = suffix
        self.include_extensions = include_extensions
        self.exclude_extensions = exclude_extensions

    def rename_files(self):
        """
        Renames the files in the root directory based on the given criteria.
        """
        def file_handler(file_path: str):
            dir_path, filename = os.path.split(file_path)
            name, extension = os.path.splitext(filename)
            new_name = f"{name}{self.suffix}{extension}"
            os.rename(file_path, os.path.join(dir_path, new_name))

        traverser = UniTraverser(self.root_dir, include_extensions=self.include_extensions,
                                 exclude_extensions=self.exclude_extensions)
        traverser.traverse(file_handler)

@click.command()
@click.argument('root_dir')
@click.argument('suffix')
@click.option('--include', '-i', default=None, type=str,
              help='File extensions to include (e.g., ".txt .jpg .webp").')
@click.option('--exclude', '-e', default=None, type=str,
              help='File extensions to exclude (e.g., ".txt .jpg .webp").')
def rename(root_dir, suffix, include, exclude):
    include_extensions = include.split() if include else None
    exclude_extensions = exclude.split() if exclude else None
    renamer = FileRenamer(root_dir, suffix, include_extensions, exclude_extensions)
    renamer.rename_files()
    print(f'Renamed files in {root_dir} with suffix {suffix}')

if __name__ == '__main__':
    rename()



import pandas as pd
import warnings
from pathlib import Path
from typing import Union, List, Dict, Any

import unibox
from unibox import UniLogger


class UniMerger:
    def __init__(self, logger: UniLogger = None):
        self.logger = logger if logger else UniLogger(file_suffix="UniMerger")

    def _merge_dicts(self, *dicts):
        result = {}
        for d in dicts:
            if not isinstance(d, dict):
                self.logger.warning(f"Expected a dictionary, but got {type(d)}. Skipping...")
                continue

            for key, value in d.items():
                if key in result and type(result[key]) != type(value):
                    self.logger.warning(
                        f"Data type mismatch for key '{key}': {type(result[key])} vs {type(value)}. Proceeding with the merge...")
                result[key] = value
        return result

    def _merge_dataframes(self, *dfs):
        common_columns = set(dfs[0].columns)
        for df in dfs[1:]:
            common_columns &= set(df.columns)

        if not common_columns:
            raise ValueError("All dataframes must have at least one column in common for merging.")

        all_columns = set()
        for df in dfs:
            all_columns |= set(df.columns)

        missing_columns = all_columns - common_columns
        if missing_columns:
            self.logger.warning(f"Columns missing in at least one dataframe: {missing_columns}")

        result = pd.concat(dfs, ignore_index=True, sort=False)
        return result

    def merges(self, *data: Union[str, Dict, pd.DataFrame, Any]):
        loaded_data = []
        for d in data:
            if isinstance(d, str):
                loaded = unibox.loads(Path(d))
                if loaded is not None:
                    loaded_data.append(loaded)
            else:
                loaded_data.append(d)

        if all(isinstance(d, dict) for d in loaded_data):
            return self._merge_dicts(*loaded_data)
        elif all(isinstance(d, pd.DataFrame) for d in loaded_data):
            return self._merge_dataframes(*loaded_data)
        else:
            self.logger.error(
                "All inputs must be of the same type (either all dictionaries, all dataframes, or all file paths).")
            return None


def demo():
    # Usage example
    data_merger = UniMerger()

    # Merging dictionaries
    dict1 = {'a': 1, 'b': 2}
    dict2 = {'c': 3, 'd': 4}
    dict3 = {'e': 5, 'f': 6}
    merged_dicts = data_merger.merges(dict1, dict2, dict3)
    print(merged_dicts)

    # Merging dataframes
    df1 = pd.DataFrame({'filename': ['file1', 'file2'], 'A': [1, 2]})
    df2 = pd.DataFrame({'filename': ['file3', 'file4'], 'B': [3, 4]})
    df3 = pd.DataFrame({'filename': ['file5', 'file6'], 'C': [5, 6]})
    merged_dfs = data_merger.merges(df1, df2, df3)
    print(merged_dfs)



if __name__ == "__main__":

    # demo()

    file1 = r"D:\Andrew\Downloads\combined_upscale_todo_v2_part5.parquet"
    file2 = r"D:\Andrew\Downloads\combined_upscale_todo_v2_part3.parquet"
    merger = UniMerger()
    df3 = merger.merges(file1, file2)

    print("D")










import concurrent.futures
import multiprocessing
import os
from pathlib import Path
from PIL import Image
from tqdm.auto import tqdm

try:
    import pyvips

    # Try to create a small image using libvips to check if it's working.
    pyvips.Image.black(1, 1)
    HAS_PYVIPS = True
except ImportError:
    HAS_PYVIPS = False
except pyvips.error.Error:
    HAS_PYVIPS = False

SUPPORTED_FORMATS = {"jpg", "webp", "png"}


# Updated UniResizer class

"""
@@@@@@@@@@@@@@@@

DEPRECIATED: use uni_resizer.py instead

@@@@@@@@@@@@@@@@
"""

class ImageResizer:
    def __init__(self, src_dir: str, dst_dir: str, min_side: int = 768, format: str = "webp", quality: int = 95,
                 exist_ok: bool = False):
        """
        Initialize an instance of UniResizer.
        """
        self.src_dir = Path(src_dir)
        self.dst_dir = Path(dst_dir)
        self.min_side = min_side
        self.format = self._validate_format(format)
        self.quality = quality
        self.exist_ok = exist_ok
        self.num_processes = multiprocessing.cpu_count()

    def _validate_format(self, format: str) -> str:
        """
        Validate the provided image format.
        """
        if format not in SUPPORTED_FORMATS:
            raise ValueError(f"Unsupported image format '{format}'. Supported formats are {SUPPORTED_FORMATS}.")
        return format

    def _resize_image(self, file_path: Path, relative_path: Path):
        """
        Resize a single image and saves the result.
        """
        # dst_path = self.dst_dir / relative_path.with_suffix(f".{self.format}")

        # new:
        rel_dst_path = relative_path.with_name(relative_path.stem + "_resized").with_suffix(f".{self.format}")

        # Ensure the destination directory exists
        rel_dst_path.parent.mkdir(parents=True, exist_ok=True)

        if HAS_PYVIPS:
            image = self._resize_with_pyvips(file_path)
        else:
            image = self._resize_with_pil(file_path)

        # Image might not need resizing but still needs to be saved in new format
        if image is None:
            if HAS_PYVIPS:
                image = pyvips.Image.new_from_file(str(file_path), access="sequential")
            else:
                image = Image.open(file_path)

        self._save_image(image, rel_dst_path)

    def _resize_with_pil(self, image_path: Path) -> Image:
        """
        Resize an image using PIL.
        """
        try:
            image = Image.open(image_path)
        except IOError:
            print(f"Error opening image {image_path}. Skipping...")
            return None

        width, height = image.size
        min_dim = min(width, height)

        if min_dim <= self.min_side:
            return None

        scale = self.min_side / min_dim
        new_width = round(width * scale)
        new_height = round(height * scale)

        return image.resize((new_width, new_height), Image.LANCZOS)

    def _resize_with_pyvips(self, image_path: Path):
        """
        Resize an image using pyvips.
        """
        try:
            image = pyvips.Image.new_from_file(str(image_path), access="sequential")
        except pyvips.error.Error:
            print(f"Error opening image {image_path} with pyvips. Skipping...")
            return None

        min_dim = min(image.width, image.height)

        if min_dim <= self.min_side:
            return None

        scale = self.min_side / min_dim
        return image.resize(scale)

    def _task(self, file_path, relative_path):
        self._resize_image(file_path, relative_path)

    def _save_image(self, image, dst_path: Path):
        """
        Save an image in the specified format.
        :param image: PIL.Image or pyvips.Image
        :param dst_path: Path (route/new_name.webp)
        """
        """
        Save an image in the specified format.
        """
        save_methods = {
            "jpg": {"pil": "JPEG", "pyvips": "jpegsave", "params": {"Q": self.quality}},
            "webp": {"pil": "WEBP", "pyvips": "webpsave", "params": {"Q": self.quality}},
            "png": {"pil": "PNG", "pyvips": "pngsave", "params": {}}
        }

        method_map = save_methods[self.format]
        try:
            if HAS_PYVIPS:
                getattr(image, method_map["pyvips"])(str(dst_path), **method_map["params"])
            else:
                image.saves(dst_path, method_map["pil"], **method_map["params"])
        except Exception as e:
            print(f"Error saving image {dst_path}. Skipping...\n{str(e)}")

    def _execute_resize(self, tasks):
        """
        Execute tasks using ProcessPoolExecutor
        """
        with concurrent.futures.ProcessPoolExecutor(max_workers=self.num_processes) as executor:
            futures = [executor.submit(task, *args) for task, *args in tasks]
            list(tqdm(concurrent.futures.as_completed(futures), total=len(tasks),
                      desc=f"Resizing images to min_side={self.min_side}"))


    def resize_single_image(self, image_path: str):
        """
        Resize a single image and saves the result next to the original image with "_resized" in the name.
        """
        file_path = Path(image_path)
        relative_path = file_path.name
        # Add "_resized" before the file extension
        resized_name = relative_path.stem + "_resized" + relative_path.suffix
        relative_resized_path = file_path.parent / resized_name

        self._resize_image(file_path, relative_resized_path)


    def resize_images(self):
        """
        Resize all supported images in the source directory.
        """
        tasks = []

        self._print_debug_string()

        for file_path in self.src_dir.rglob("*"):
            if file_path.is_file() and file_path.suffix[1:].lower() in SUPPORTED_FORMATS:
                relative_path = file_path.relative_to(self.src_dir)
                dst_path = self.dst_dir / relative_path.with_suffix(f".{self.format}")

                # Check if the destination file exists and exist_ok is True
                if not self.exist_ok or not dst_path.exists():
                    tasks.append((self._task, file_path, relative_path))

        self._execute_resize(tasks)

    def _print_debug_string(self):
        """
        prints debug message once resize_images is called
        """
        debug_string = ""
        if HAS_PYVIPS:
            debug_string += "Resizing with libvips: "
        else:
            debug_string += "Resizing with PIL: "
        debug_string += f"num_processes = {self.num_processes}, src_dir={self.src_dir}, dst_dir={self.dst_dir}, " \
                        f"min_side={self.min_side}, format={self.format}, quality={self.quality}," \
                        f"exist_ok={self.exist_ok}"
        print(debug_string)


if __name__ == "__main__":
    # Define the source and destination directories and the minimum side length for resizing
    src_dir = input("s3_uri path:")

    dst_dir = f"{src_dir}_768webp"
    min_side = 1024

    # Create an instance of the UniResizer class
    resizer = ImageResizer(src_dir, dst_dir, min_side)

    # Get a list of image files from the source directory
    resizer.resize_images()



import os
import click

import unibox
from .processing import image_resizer
from .processing import file_mover


@click.group()
def cli():
    pass


@cli.command()
@click.argument('src_dir')
@click.option('--min_side', '-m', default=None, type=int, help='Minimum side length of the image.')
@click.option('--dst_dir', '-d', default=None, help='Destination directory for the images.')
@click.option('--format', '-f', default=None, help='Format of the image.')
@click.option('--quality', '-q', default=None, type=int, help='Quality of the image.')
@click.option('--exist_ok', '-e', default=None, type=bool, help='Keep existing images')
def resize_legacy(src_dir, min_side, dst_dir, format, quality, exist_ok):
    # resize implementation
    print(f'Resizing directory: {src_dir}')

    if not min_side:
        min_side = click.prompt('- Minimum side length', type=int, default=768)
    if not format:
        format = click.prompt('- File format', default="webp")
    if not dst_dir:
        src_dir_name = os.path.basename(os.path.abspath(src_dir))
        dst_dir = click.prompt('- Destination', default=f"{src_dir_name}_{min_side}_{format}")
    if not quality:
        quality = click.prompt('- Image quality', type=int, default=95)
    if not exist_ok:
        exist_ok = click.prompt('- Keep existing images', type=bool, default=True)

    # Printing the command
    print("\nRunning Command:\n")
    print(f"unibox resize {src_dir} --min_side {min_side} --dst_dir {dst_dir} "
          f"--format {format} --quality {quality} --exist_ok {exist_ok}\n")

    resizer = image_resizer.ImageResizer(src_dir, dst_dir, min_side, format=format, quality=quality, exist_ok=exist_ok)
    resizer.resize_images()


from unibox import UniResizer


@cli.command()
@click.option("--img_root_dir", "-d", default=None, type=str, help="Root directory of the images.")
@click.option("--to_dir", "-t", default=None, type=str, help="Target directory for the images.")
@click.option("--min_side", "-m", default=None, type=int, help="Minimum side length of the image.")
@click.option("--max_side", "-M", default=None, type=int, help="Maximum side length of the image.")
@click.option("--target_pixels", "-p", default=None, type=int,
              help="Target number of pixels of the image.")
@click.option("--keep_hierarchy", "-k", default=True, type=bool, help="Keep directory hierarchy.")
@click.option("--exist_ok", "-e", default=True, type=bool, help="Keep existing images.")
def resize(img_root_dir, to_dir, min_side, max_side, target_pixels, keep_hierarchy, exist_ok):
    if not img_root_dir:
        img_root_dir = click.prompt('- Image root directory', type=str)

    if not target_pixels:
        target_pixels = click.prompt('- Target number of pixels (-1 to unset)', type=int,
                                     default=int(1024 * 1024 * 1.25))
        if target_pixels == -1:
            target_pixels = None

    dir_str = ""
    if min_side:
        dir_str += f"_min{min_side}"
    if max_side:
        dir_str += f"_max{max_side}"
    if target_pixels:
        dir_str += f"_target{target_pixels}"

    if not to_dir:
        src_dir_name = os.path.basename(os.path.abspath(img_root_dir))
        to_dir = click.prompt('- Destination', default=f"{src_dir_name}_resized{dir_str}_webp")

    if not keep_hierarchy:
        keep_hierarchy = click.prompt('- Keep directory hierarchy', type=bool, default=True)

    if not exist_ok:
        exist_ok = click.prompt('- Keep existing images', type=bool, default=True)

    resizer = UniResizer(img_root_dir, to_dir,
                         min_dim=min_side, max_dim=max_side, target_pixels=target_pixels,
                         keep_hierarchy=keep_hierarchy, exist_ok=exist_ok)

    resizer.execute_resize_jobs(resizer.get_resize_jobs())


def extensions_option(value: str):
    if value:
        extensions = value.strip().split()
        return [f".{ext.strip()}" if not ext.startswith('.') else ext.strip() for ext in extensions]
    return None


@cli.command()
@click.argument('src_dir')
@click.argument('dst_dir')
@click.option('--keep_structure', '-k', is_flag=True, help='Keep directory structure.')
@click.option('--include', '-i', default=None, type=extensions_option,
              help='File extensions to include (e.g., ".txt .jpg .webp").')
@click.option('--exclude', '-e', default=None, type=extensions_option,
              help='File extensions to exclude (e.g., ".txt .jpg .webp").')
def move(src_dir, dst_dir, keep_structure, include, exclude):
    mover = file_mover.DirMoverCopier(src_dir, dst_dir, keep_structure, include, exclude, 'move')
    mover.process_files()
    print(f'Moved files from {src_dir} to {dst_dir}')


@cli.command()
@click.argument('src_dir')
@click.argument('dst_dir')
@click.option('--keep_structure', '-k', is_flag=True, help='Keep directory structure.')
@click.option('--include', '-i', default=None, type=extensions_option,
              help='File extensions to include (e.g., ".txt .jpg .webp").')
@click.option('--exclude', '-e', default=None, type=extensions_option,
              help='File extensions to exclude (e.g., ".txt .jpg .webp").')
def copy(src_dir, dst_dir, keep_structure, include, exclude):
    copier = file_mover.DirMoverCopier(src_dir, dst_dir, keep_structure, include, exclude, 'copy')
    copier.process_files()
    print(f'Copied files from {src_dir} to {dst_dir}')


if __name__ == '__main__':
    cli()



import os
import math
from PIL import Image
from pathlib import Path
from tqdm.auto import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import List, Tuple
import unibox
from unibox import UniLogger


def chunk_list(lst, chunk_size):
    """Yield successive chunks of size `chunk_size` from list `lst`."""
    for i in range(0, len(lst), chunk_size):
        yield lst[i:i + chunk_size]


class UniResizer:
    TRUNCATE_MULTIPLIER = 32  # the shorter side will be a multiple of 32 after scaling
    WEBP_QUALITY = 98  # quality of the output image (max:100)
    EXTENSION = ".webp"  # file extension (.webp)

    def __init__(self, root_dir: str, dst_dir: str, min_dim: int = None, max_dim: int = None, target_pixels: int = None,
                 keep_hierarchy: bool = True, exist_ok: bool = True, logger: UniLogger = None, max_workers: int = None,
                 chunk_size: int = 10_000, ):
        """
        Resizes images in a directory to a specified size.

        Resolution:
            - by min_dim: resize the shorter side to a specified size
            - by max_dim: resize the longer side to a specified size
            - by target_pixels: resize the shorter side to a specified number of pixels

        Methods:
            - by directory: resize all images in a directory
            - by list: resize a list of images given

        Configs:
            - keep_hierarchy: keep the same directory structure as the source directory
            - exist_ok: skip resizing if the destination file already exists
            - max_workers: number of workers for parallel processing
            - chunk_size: number of images to process in each chunk (to be dispatched to ProcessPoolExecutor each time)
        
        Usage:
        >>> resizer = UniResizer(root_dir, dst_dir, min_dim=min_dim, max_dim=max_dim, target_pixels=target_pixels,
        >>>                   keep_hierarchy=False, exist_ok=True)
        >>> resizer.execute_resize_jobs(resizer.get_resize_jobs())
        """ 
        
        self.root_dir = root_dir
        self.dst_dir = dst_dir
        self.min_dim = min_dim
        self.max_dim = max_dim
        self.target_pixels = target_pixels
        self.keep_hierarchy = keep_hierarchy
        self.exist_ok = exist_ok
        self.logger = logger if logger is not None else UniLogger()
        self.max_workers = int(os.cpu_count() - 1) if max_workers is None else max_workers
        self.chunk_size = chunk_size

    def _get_dst_path(self, og_rel_image_path: str) -> str:
        og_rel_path = Path(og_rel_image_path)
        rel_resized_path = og_rel_path.with_name(f"{og_rel_path.stem}_resized{self.EXTENSION}")
        if self.keep_hierarchy:
            dst_file_path = Path(self.dst_dir) / rel_resized_path
        else:
            dst_file_path = Path(self.dst_dir) / rel_resized_path.name
        return str(dst_file_path)

    def _create_dst_dir(self, dst_file_path: str) -> None:
        Path(dst_file_path).parent.mkdir(parents=True, exist_ok=True)

    def resize_single_image(self, image: Image.Image) -> Image.Image:
        width, height = image.size
        new_width, new_height = width, height

        need_resize = (self.min_dim is not None and min(width, height) < self.min_dim) or \
                      (self.max_dim is not None and max(width, height) > self.max_dim)

        if self.target_pixels is not None:
            if width * height > self.target_pixels:
                scale_factor = math.sqrt(self.target_pixels / (width * height))
                target_shorter_side = round((min(width, height) * scale_factor) // self.TRUNCATE_MULTIPLIER) * self.TRUNCATE_MULTIPLIER
                new_width, new_height = self._get_new_dimensions(width, height, target_shorter_side)

        elif self.max_dim is not None and need_resize:
            new_width, new_height = self._get_new_dimensions(width, height, self.max_dim, resize_by_longer_side=True)

        elif self.min_dim is not None and need_resize:
            new_width, new_height = self._get_new_dimensions(width, height, self.min_dim)

        elif not need_resize:
            return image

        return image.resize((new_width, new_height), Image.LANCZOS)

    @staticmethod
    def _get_new_dimensions(width: int, height: int, target_side: int, resize_by_longer_side: bool = False) -> tuple:
        if target_side == -1:
            return int(width), int(height)
        if any(x <= 0 or type(x) != int for x in [width, height, target_side]):
            raise ValueError("width, height and target_side must be positive integers")
        shorter_side, longer_side = min(width, height), max(width, height)
        if resize_by_longer_side:
            new_longer_side = target_side
            new_shorter_side = round(new_longer_side * (shorter_side / longer_side))
        else:
            new_shorter_side = target_side
            new_longer_side = round(new_shorter_side * (longer_side / shorter_side))
        if width > height:
            return new_longer_side, new_shorter_side
        else:
            return new_shorter_side, new_longer_side

    def _resize_single_image_task(self, og_rel_image_path: str) -> None:
        src_file_path = os.path.join(self.root_dir, og_rel_image_path)
        dst_file_path = self._get_dst_path(og_rel_image_path)
        if self.keep_hierarchy:
            self._create_dst_dir(dst_file_path)
        try:
            with Image.open(src_file_path) as image:
                resized_image = self.resize_single_image(image)
                with open(dst_file_path, "wb") as f:
                    resized_image.save(f, "webp", quality=self.WEBP_QUALITY)
        except OSError as e:
            self.logger.error(f"Error processing image {og_rel_image_path}: {e}. Skipping...")

    @staticmethod
    def _execute_resize_tasks(tasks: List[Tuple], max_workers: int, total_progress: tqdm) -> None:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            future_to_task = {executor.submit(task, *args): i for i, (task, *args) in enumerate(tasks)}
            for future in as_completed(future_to_task):
                total_progress.update(1)

    def get_resize_jobs(self) -> List[str]:
        self.logger.info("Getting image paths...")
        todo_image_files = unibox.traverses(self.root_dir, include_extensions=unibox.constants.IMG_FILES, relative_unix=True)
        if self.exist_ok:
            self.logger.info("Checking existing files...")
            existing_files = unibox.traverses(self.dst_dir, include_extensions=unibox.constants.IMG_FILES, relative_unix=False)
            expected_file_set = {self._get_dst_path(rel_path) for rel_path in todo_image_files}
            existing_file_set = set(existing_files)
            needed_files_set = expected_file_set - existing_file_set
            reverse_mapping = {self._get_dst_path(rel_path): rel_path for rel_path in todo_image_files}
            todo_image_files = [reverse_mapping[dst_path] for dst_path in needed_files_set]
        return todo_image_files

    def execute_resize_jobs(self, image_files: List[str], report_interval:float=5) -> None:
        tasks = [(self._resize_single_image_task, og_rel_image_path) for og_rel_image_path in image_files]
        self.logger.info(f"Resizing {len(tasks)} images from {self.root_dir} to {self.dst_dir} | chunk size: {self.chunk_size} | report interval: {report_interval}s")
        
        # creating shared progress bar
        pbar = tqdm(total=len(tasks), 
                    desc=f"[interval={report_interval}] resizing [{self.root_dir}]", 
                    mininterval=report_interval
                    )
        # Process each chunk
        for chunk in chunk_list(tasks, self.chunk_size):
            self._execute_resize_tasks(chunk, self.max_workers, pbar)
        
        pbar.close()

# Example Usage
if __name__ == '__main__':
    root_dir = r"E:\_benchmark\1k"
    dst_dir = r"E:\_benchmark\1k_resized"
    min_dim = 512
    max_dim = int(1024 * 3)
    target_pixels = int(1024 * 1024 * 1.25)
    resizer = UniResizer(root_dir, dst_dir, min_dim=min_dim, max_dim=max_dim, target_pixels=target_pixels, keep_hierarchy=False, exist_ok=True)
    images_to_resize = resizer.get_resize_jobs()
    resizer.execute_resize_jobs(images_to_resize)



"commonly used constant variables"

__all_imgs_raw = (
    "jpg jpeg png bmp dds exif jp2 jpx pcx pnm ras gif tga tif tiff xbm xpm webp jpe"
)
IMG_FILES = ["." + i.strip() for i in __all_imgs_raw.split(" ")]

# new name
IMAGE_FILES = IMG_FILES



import os
import inspect
import logging
import colorlog

from pathlib import Path
from datetime import datetime

NOTICE = 25  # Value between WARNING (30) and INFO (20)


class UniLogger:
    """包含时间和 caller frame的 logger utils

    VERSION: 2023.06.11 (JUNE)

    >>> logger = UniLogger("test", logger_name=__name__)
    >>> logger.info("test")


    (Use in a utils:)

    if logger is not None:
        self.logger = logger
    else:
        logging.basicConfig(level=logging.INFO,
                            format='%(asctime)s [%(levelname)s] %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S')
        self.logger = logging.getLogger(__name__)
    """

    def __init__(self, output_dir: str = "logs", file_suffix: str = "log", verbose: bool = True,
                 logger_name: str = None, write_log: bool = True):
        """
        Initialize an instance of UniLogger.

        :param output_dir: directory to saves the log file
        :param file_suffix: suffix of the log file
        :param verbose: if True, set the log level to DEBUG; otherwise, set to INFO
        :param logger_name: name of the logger
        :param write_log: if True, write the log to a file; otherwise, only print to console
        """

        self.verbose = verbose
        self.write_log = write_log

        handlers = []

        if self.write_log:
            self.output_dir = Path(output_dir)
            self.log_file_suffix = file_suffix
            self.output_dir.mkdir(parents=True, exist_ok=True)
            self.log_file = (
                    self.output_dir
                    / f"{self.log_file_suffix}_{datetime.now().strftime('%Y%m%d')}.log"
            )
            handlers.append(logging.FileHandler(self.log_file, mode="a", encoding="utf-8"))

        # Create a color handler
        color_handler = colorlog.StreamHandler()
        color_handler.setFormatter(
            colorlog.ColoredFormatter(
                fmt='%(log_color)s%(asctime)s [%(levelname)s] %(name)s: %(message)s',
                datefmt='%Y-%m-%d %H:%M:%S',
                log_colors={
                    'DEBUG': 'cyan',
                    'INFO': 'green',
                    'WARNING': 'yellow',
                    'ERROR': 'red',
                    'CRITICAL': 'red,bg_white',
                },
            )
        )
        handlers.append(color_handler)

        if verbose:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO

        self.logger = logging.getLogger(logger_name if logger_name else self.__class__.__name__)
        self.logger.setLevel(log_level)

        # Check if handlers already exist before adding them; prevents duplicate handlers in Jupyter notebooks
        if not self.logger.hasHandlers():
            formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s")
            for handler in handlers:
                handler.setFormatter(formatter)
                self.logger.addHandler(handler)

        logging.addLevelName(NOTICE, "NOTICE")

    def log(self, log_level: str, message: str):
        level = getattr(logging, log_level.upper(), logging.INFO)

        # Get the caller function information
        caller_frame = inspect.currentframe().f_back.f_back  # Skip 'log' function frame
        caller_func_name = caller_frame.f_code.co_name

        # Check if called from a utils
        if "self" in caller_frame.f_locals:
            caller_class_name = caller_frame.f_locals["self"].__class__.__name__
        else:
            caller_class_name = None

        # Include the caller function name in the message
        if caller_class_name:
            full_message = f"{caller_class_name}.{caller_func_name}: {message}"
        else:
            full_message = f"{caller_func_name}: {message}"

        self.logger.log(level, full_message)

    def notice(self, message: str):
        """
        using INFO log level but with a checkmark √
        """
        self.log("NOTICE", "✅ " + message)

    def warning(self, message: str):
        self.log("WARNING", "⚠️ " + message)

    def error(self, message: str):
        self.log("ERROR", "❌ " + message)

    def info(self, message: str):
        self.log("INFO", message)
